/**********************************************************
**
**		Kinematic Test
**		Copper / Defect Softworks
**		02.05.2001  
**		24.05.2001
**		16.11.2001 ported to PalmOS by Thomas Runge
*/

#include <PalmOS.h>
#include <handera/vga.h>
#include "resource.h"
#include "mathutil.h"
#include "kine.h"
#ifdef USE_MATHLIB
#include "MathLib.h"
#else
extern double sqrt(double x);
#endif

static int StartApplication(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean FormHandler(EventPtr event);
static void HideObjects(FormPtr form, Boolean hide);
static void HideObject(FormPtr form, UInt16 id, Boolean hide);
static void MoveObjects(FormPtr form, Coord y_diff);
static void MoveObject(FormPtr form, UInt16 objIndex, Coord y_diff);
static void ShowMessageDialog(UInt16 fid);
static int OnPenDown(Coord x, Coord y);
static int OnPenMove(Coord x, Coord y);
static int OnPenUp(Coord x, Coord y);
static int OnKey(WChar key);
static void OnResize(FormPtr form);
static void CDrawLine(Coord x1, Coord y1, Coord x2, Coord y2,
							IndexedColorType color);
static void CDrawRectangle(Coord x, Coord y, Coord w, Coord h,
							IndexedColorType color);
static void drawjoints();
static void makejoints();
static void GetJacobi(GenMatrix *A);
static void CalcNewAngel(Int32 dx, Int32 dy);
static void MakeNewPosition(Coord xx, Coord yy);
static void* GetObjectPtr(FormPtr form, UInt16 objectID);

#define XWIN 200
#define YWIN 150

world *w;
Coord width, height;
Boolean has_vga = false;
Boolean has_color = false;
Boolean is_testing = false;
RectangleType clip;
UInt16 title_height;

static const RGBColorType _dk_blue = {0x00, 0x00, 0x00, 0xFF};
static const RGBColorType _lt_blue = {0x00, 0x55, 0x55, 0xFF};
static const RGBColorType _red     = {0x00, 0xDD, 0x00, 0x00};

IndexedColorType dk_blue;
IndexedColorType lt_blue;
IndexedColorType red;

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	int error;

	if(cmd == sysAppLaunchCmdNormalLaunch)
	{
		error = StartApplication();
		if(error)
			return(error);
		EventLoop();
		StopApplication();
	}
	return(0);
}

static int StartApplication(void)
{
	Err err;
	UInt32 version;

	w = MemPtrNew(sizeof(world));
	if(w == NULL)
		return(1);

#ifdef USE_MATHLIB
	err = SysLibFind(MathLibName, &MathLibRef);
	if(err)
		err = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
	if(err)
	{
		ErrFatalDisplayIf(1, "Can't find MathLib");
		return(1);
	}
	err = MathLibOpen(MathLibRef, MathLibVersion);
	if(err)
	{
		ErrFatalDisplayIf(1, "Can't open MathLib"); 
		return(1);
	}
#endif

	err = WinScreenMode(winScreenModeGet, NULL, NULL, NULL, &has_color);
	if(err)
	{
		ErrDisplay("Failed to get screen mode");
		return(1);
	}

	if(has_color)
	{
		dk_blue = WinRGBToIndex(&_dk_blue);
		lt_blue = WinRGBToIndex(&_lt_blue);
		red     = WinRGBToIndex(&_red);
	}

	if(FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0)
	{
		has_vga = true;
		VgaSetScreenMode(screenMode1To1, rotateModeNone);
		title_height = VgaGetFrmTitleHeight()+1;
	}
	else
		title_height = 16;

	w->kc[0].j.theta = 120 * PI / 180;
	w->kc[0].j.d = 0;
	w->kc[0].j.a = 30;
	w->kc[0].j.a1 = 30;
	w->kc[0].j.alpha = 0;

	w->kc[1].j.theta = 280 * PI / 180;
	w->kc[1].j.d = 0;
	w->kc[1].j.a = 40;
	w->kc[1].j.a1 = 40;
	w->kc[1].j.alpha = 0;

	w->kc[2].j.theta = 310 * PI / 180;
	w->kc[2].j.d = 0;
	w->kc[2].j.a = 30;
	w->kc[2].j.a1 = 30;
	w->kc[2].j.alpha = 0;

	w->kc[3].j.theta = 0 * PI / 180;
	w->kc[3].j.d = 0;
	w->kc[3].j.a = 15;
	w->kc[3].j.a1 = 15;
	w->kc[3].j.alpha = 0;

	w->length = 200.;
	w->baselength = 100;
	w->basex = XWIN / 2 - w->baselength / 2;
	w->basey = YWIN/3;
	w->xoff = XWIN / 2;
	w->yoff = w->basey;

	FrmGotoForm(KineForm);
	return(0);
}

static void StopApplication(void)
{
#ifdef USE_MATHLIB
	UInt16 usecount;
	Err err;

	err = MathLibClose(MathLibRef, &usecount);
	ErrFatalDisplayIf(err, "Can't close MathLib");
	if(usecount == 0)
		SysLibRemove(MathLibRef); 
#endif
	MemPtrFree(w);
}

static void EventLoop(void)
{
	Err err;
	int formID;
	FormPtr form;
	EventType event;

	do
	{
		EvtGetEvent(&event, -1);

		if(SysHandleEvent(&event))
			continue;

		if(MenuHandleEvent((void *)0, &event, &err))
			continue;

		if(event.eType == frmLoadEvent)
		{
			formID = event.data.frmLoad.formID;
			form = FrmInitForm(formID);
			FrmSetActiveForm(form);
			switch(formID)
			{
				case KineForm:
						FrmSetEventHandler(form, (FormEventHandlerPtr)FormHandler);
				break;
			}
		}
		FrmDispatchEvent(&event);
	}
	while(event.eType != appStopEvent);
}

static Boolean FormHandler(EventPtr event)
{
	FormPtr	form = FrmGetActiveForm();
	int		handled = 0;
	UInt32	kc_num	= 0;
	Int8	sign	= 1;

	switch(event->eType)
	{
		case frmOpenEvent:
			if(has_vga)
				VgaFormModify(form, vgaFormModify160To240);
			OnResize(form);
			handled = 1;
			break;
		case displayExtentChangedEvent:
			OnResize(form);
			handled = 1;
			break;
		case penDownEvent:
			handled = OnPenDown(event->screenX, event->screenY);
			break;
		case penMoveEvent:
			handled = OnPenMove(event->screenX, event->screenY);
			break;
		case penUpEvent:
			handled = OnPenUp(event->screenX, event->screenY);
			break;
		case keyDownEvent:
			handled = OnKey(event->data.keyDown.chr);
			break;
		case ctlRepeatEvent:
			switch(event->data.ctlRepeat.controlID)
			{
				case left4ID:
					kc_num = 3;
					sign   = 1;
					break;
				case right4ID:
					kc_num = 3;
					sign   = -1;
					break;
				case left3ID:
					kc_num = 2;
					sign   = 1;
					break;
				case right3ID:
					kc_num = 2;
					sign   = -1;
					break;
				case left2ID:
					kc_num = 1;
					sign   = 1;
					break;
				case right2ID:
					kc_num = 1;
					sign   = -1;
					break;
				case right1ID:
					kc_num = 0;
					sign   = -1;
					break;
				case left1ID:
				default:
					kc_num = 0;
					sign   = 1;
					break;
			}
			w->kc[kc_num].j.theta += sign * 2 * PI / 180;
			makejoints();
			drawjoints();
			handled = 0;
			break;
		case menuEvent:
			switch(event->data.menu.itemID)
			{
				case OptionsAbout:
						ShowMessageDialog(AboutForm);
					break;
				case OptionsTest:
						is_testing = !is_testing;
						OnResize(form);
					break;
			}
			handled = 1;
			break;
		case nilEvent:
			handled = 1;
			break;
		default:
			handled = 0;
			break;
	}
	return(handled);
}

static void HideObjects(FormPtr form, Boolean hide)
{
	HideObject(form, left1ID, hide);
	HideObject(form, right1ID, hide);
	HideObject(form, left2ID, hide);
	HideObject(form, right2ID, hide);
	HideObject(form, left3ID, hide);
	HideObject(form, right3ID, hide);
	HideObject(form, left4ID, hide);
	HideObject(form, right4ID, hide);
}

static void HideObject(FormPtr form, UInt16 id, Boolean hide)
{
	UInt16 objID = FrmGetObjectIndex(form, right1ID);
	if(hide)
		FrmHideObject(form, objID);
	else
		FrmShowObject(form, objID);
}

static void MoveObjects(FormPtr form, Coord y_diff)
{
	MoveObject(form, FrmGetObjectIndex(form, left1ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, right1ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, left2ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, right2ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, left3ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, right3ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, left4ID), y_diff);
	MoveObject(form, FrmGetObjectIndex(form, right4ID), y_diff);
}

static void MoveObject(FormPtr form, UInt16 objIndex, Coord y_diff)
{
    RectangleType r;
    
    FrmGetObjectBounds(form, objIndex, &r);
    r.topLeft.y += y_diff;
    FrmSetObjectBounds(form, objIndex, &r);
}

static void ShowMessageDialog(UInt16 fid)
{
	FormPtr form;

	form = FrmInitForm(fid);
	if(form)
	{
		if(has_vga)
			VgaFormModify(form, vgaFormModify160To240);
		FrmDoDialog(form);
		FrmDeleteForm(form);
	}
}

static int OnPenDown(Coord x, Coord y)
{
	if(!RctPtInRectangle(x, y, &clip))
		return(0);

	w->check = false;

	if(ABS(x - w->ex - w->xoff) <= 4)
		if(ABS(height - y - w->ey - w->yoff) <= 4)
			w->check = true;

	if(!w->check)
		MakeNewPosition(x, y);
	return(1);
}

static int OnPenMove(Coord x, Coord y)
{
	if(!RctPtInRectangle(x, y, &clip))
		return(0);

	if(!w->check)
		MakeNewPosition(x, y);
	return(1);
}

static int OnPenUp(Coord x, Coord y)
{
	if(!RctPtInRectangle(x, y, &clip))
		return(0);

	w->check = false;
	return(1);
}

static int OnKey(WChar key)
{
	return(0);
}

static void OnResize(FormPtr form)
{
	float sc;
	Int32 i;
	RectangleType old_size;
	Coord y_diff;

	WinGetDisplayExtent(&width, &height);
	FrmGetFormBounds(form, &old_size);
	y_diff = height - (old_size.topLeft.y + old_size.extent.y);

	clip.topLeft.x = 0;
	clip.topLeft.y = 0;
	clip.extent.x = width;
	clip.extent.y = height;
	WinSetWindowBounds(FrmGetWindowHandle(form), &clip);
	clip.extent.y = (Coord) (height - title_height);
	clip.topLeft.y = title_height;

	if(y_diff)
		MoveObjects(form, y_diff);

	if(!is_testing)
		HideObjects(form, true);
	else
	{
		FieldPtr field;
		RectangleType bounds;

		HideObjects(form, false);
		field = GetObjectPtr(form, left1ID);
		FldGetBounds(field, &bounds);
		clip.extent.y -= bounds.extent.y+3;

		WinEraseWindow();
	}
	FrmDrawForm(form);

	sc = (float) (sqrt((float)width * (float)height) /
					sqrt((float)XWIN * (float)YWIN));

	w->length = 0.;
	for(i=0;i<JMax;i++)
	{
		w->kc[i].j.a = (Int32) ((float) w->kc[i].j.a1 * sc);
		w->length += (float)w->kc[i].j.a;
	}

	w->baselength = 100 * sc;
	w->basex = (Int32) ((float)width / 2. - (float)w->baselength / 2.);
	w->basey = (Int32) ((float)height / 3.);
	w->xoff = (Int32) ((float)width / 2.);
	w->yoff = (Int32) w->basey;

	makejoints();
	drawjoints();
}

static void CDrawLine(Coord x1, Coord y1, Coord x2, Coord y2,
							IndexedColorType color)
{
	IndexedColorType old_color = 1;

	if(has_color)
		old_color = WinSetForeColor(color);
	WinDrawLine(x1, y1, x2, y2);
	if(has_color)
		WinSetForeColor(old_color);
}

static void CDrawRectangle(Coord x, Coord y, Coord w, Coord h,
							IndexedColorType color)
{
	RectangleType rect;
	IndexedColorType old_color = 1;

	if(has_color)
		old_color = WinSetForeColor(color);

	rect.topLeft.x = x;
	rect.topLeft.y = y;
	rect.extent.x = w;
	rect.extent.y = h;
	WinDrawRectangle(&rect, 0);
	if(has_color)
		WinSetForeColor(old_color);
}

/*
**
** Draw Joints
**
*/
static void drawjoints()
{
	Int32 xoff = w->xoff;
	Int32 yoff = w->yoff;
	Int32 tx,ty,tx1,ty1;

	WinSetClip(&clip);
	WinEraseWindow();

	CDrawLine(w->basex,height-w->basey,w->basex+w->baselength,height-w->basey,
				dk_blue);
	CDrawRectangle(xoff-2, height-yoff-2, 4, 4, red);

	tx1 = (Int32)w->kc[0].px+xoff;
	ty1 = (Int32)w->kc[0].py+yoff;

	tx = (Int32)w->kc[1].px+xoff;
	ty = (Int32)w->kc[1].py+yoff;

	CDrawLine(tx1,height-ty1,tx,height-ty, lt_blue);
	CDrawRectangle(tx-2,height-ty-2,4,4, red);

	tx1 = (Int32)w->kc[2].px+xoff;
	ty1 = (Int32)w->kc[2].py+yoff;

	CDrawLine(tx,height-ty,tx1,height-ty1, lt_blue);
	CDrawRectangle(tx1-2,height-ty1-2,4,4, red);

	tx = (Int32)w->kc[3].px+xoff;
	ty = (Int32)w->kc[3].py+yoff;

	CDrawLine(tx1,height-ty1,tx,height-ty, lt_blue);
	CDrawRectangle(tx-2,height-ty-2,4,4, red);

	tx1 = (Int32)w->ex+xoff;
	ty1 = (Int32)w->ey+yoff;

	CDrawLine(tx,height-ty,tx1,height-ty1, lt_blue);
	CDrawRectangle(tx1-4,height-ty1-4,6,6, red);
	WinResetClip();
}

/* 
** Berechne Gelenkpositionen mit Hartenbergtrafo
*/
static void makejoints()
{
	/* Erster Punkt */
	w->v[0] = 0;
	w->v[1] = 0;
	w->v[2] = 0;
	w->v[3] = 1;

	w->kc[0].px = 0;
	w->kc[0].py = 0;
	w->kc[0].pz = 0;

	GetHartenberg(&w->kc[0].j, &w->kc[0].m);

	MatMultPoint(&w->kc[0].m , w->v, w->tmp);
	w->kc[1].px = w->tmp[0];
	w->kc[1].py = w->tmp[1];
	w->kc[1].pz = w->tmp[2];

	GetHartenberg(&w->kc[1].j, &w->mat1);

	MatMultGeneral(&w->kc[0].m,&w->mat1,&w->kc[1].m);

	MatMultPoint(&w->kc[1].m, w->v, w->tmp);
	w->kc[2].px = w->tmp[0];
	w->kc[2].py = w->tmp[1];
	w->kc[2].pz = w->tmp[2];

	GetHartenberg(&w->kc[2].j, &w->mat1);

	MatMultGeneral(&w->kc[1].m,&w->mat1,&w->kc[2].m);

	MatMultPoint(&w->kc[2].m,w->v,w->tmp);
	w->kc[3].px = w->tmp[0];
	w->kc[3].py = w->tmp[1];
	w->kc[3].pz = w->tmp[2];

	GetHartenberg(&w->kc[3].j, &w->mat1);

	MatMultGeneral(&w->kc[2].m,&w->mat1,&w->kc[3].m);
	MatMultPoint(&w->kc[3].m,w->v,w->tmp);

	w->ex = w->tmp[0];
	w->ey = w->tmp[1];
	w->ez = w->tmp[2];
}

/*
**
** 	Calculate JacobiMatrix
**
*/
static void GetJacobi(GenMatrix *A)
{
	UInt16 i,j,t;
	float bx,by,bz;
#define a(x) (w->kc[j].m.v[OF_##x])
	
	/* Copy 'a' and 'b' Coeffizents */
	
	i=0;
	j=0;
	
	for(t=0;t < JMax;t++,j++,i+=3)
	{
		bx = w->ex - w->kc[j].px;
		by = w->ey - w->kc[j].py;
		bz = w->ez - w->kc[j].pz;

		A->m[3][i] = a(11);
		A->m[4][i] = a(21);
		A->m[5][i] = a(31);

		A->m[0][i] = a(21)*bz - a(31)*by;
		A->m[1][i] = a(31)*bx - a(11)*bz;
		A->m[2][i] = a(11)*by - a(21)*bx;
		
		A->m[3][i+1] = a(12);
		A->m[4][i+1] = a(22);
		A->m[5][i+1] = a(32);

		A->m[0][i+1] = a(22)*bz - a(32)*by;
		A->m[1][i+1] = a(32)*bx - a(12)*bz;
		A->m[2][i+1] = a(12)*by - a(22)*bx;

		A->m[3][i+2] = a(13);
		A->m[4][i+2] = a(23);
		A->m[5][i+2] = a(33);

		A->m[0][i+2] = a(23)*bz - a(33)*by;
		A->m[1][i+2] = a(33)*bx - a(13)*bz;
		A->m[2][i+2] = a(13)*by - a(23)*bx;

	}
	
#undef a
}


/*
**
**  calculate the new angels for a new position
**
*/
static void CalcNewAngel(Int32 dx, Int32 dy)
{
	float x1,x2,y1,y2,tx,ty;

	InitGenMatrix(&w->gmat1,6,3*JMax);
	InitGenMatrix(&w->gmat2,3*JMax,6);
	InitGenMatrix(&w->gmat3,6,6);

	x1 = w->ex + (float)dx;
	y1 = w->ey + (float)dy;

	GetJacobi(&w->gmat1);

	//GenMatPseudoInvers(&w->gmat1,&w->gmat2,&w->gmat3);

	/* Calculate Pseudoinverse */
	/* J^T(JJ^T+kI)^-1, k damping factor to minimize singualies */
	GenMatTranspose(&w->gmat1,&w->gmat2);
	GenMatMultiply(&w->gmat1,&w->gmat2,&w->gmat3);

	InitGenMatrix(&w->gmat5,6,6);
	GenMatLoadIdentity(&w->gmat5,0.5);
	InitGenMatrix(&w->gmat4,6,6);
	GenMatAdd(&w->gmat3,&w->gmat5,&w->gmat4);
	GenMatInvers(&w->gmat4);
	w->gmat1.rows = w->gmat2.rows;
	w->gmat1.column = w->gmat2.column;
	GenMatMultiply(&w->gmat2,&w->gmat4,&w->gmat1);

	InitGenMatrix(&w->gmat4,6,1);
	w->gmat4.m[0][0] = (float)dx;
	w->gmat4.m[1][0] = (float)dy;
	w->gmat4.m[2][0] = 0.;
	w->gmat4.m[3][0] = 0.;
	w->gmat4.m[4][0] = 0.;
	w->gmat4.m[5][0] = 0.;

	InitGenMatrix(&w->gmat5,3*JMax,1);

	GenMatMultiply(&w->gmat1,&w->gmat4,&w->gmat5);

	w->kc[0].j.theta += w->gmat5.m[2][0];
	w->kc[1].j.theta += w->gmat5.m[5][0];
	w->kc[2].j.theta += w->gmat5.m[8][0];
	w->kc[3].j.theta += w->gmat5.m[11][0];

	makejoints();

	x2 = x1 - w->ex;
	y2 = y1 - w->ey;

	w->gmat4.m[0][0] = x2;
	w->gmat4.m[1][0] = y2;


	GenMatMultiply(&w->gmat1,&w->gmat4,&w->gmat5);

	w->kc[0].j.theta += w->gmat5.m[2][0];
	w->kc[1].j.theta += w->gmat5.m[5][0];
	w->kc[2].j.theta += w->gmat5.m[8][0];
	w->kc[3].j.theta += w->gmat5.m[11][0];

	makejoints();

	/* Fehler Testen */
	tx = w->ex-x1;
	ty = w->ey-y1;
}


/*
**	Calculate new Position
**
*/
static void MakeNewPosition(Coord xx, Coord yy)
{
		
	Int32 x1,y1,x2,y2,k,steps,dx,dy;
	float xincrement,yincrement,x,y;
	
	x1 = w->ex;
	y1 = w->ey;
	x2 = (Int32)xx - w->xoff;
	y2 = height - (Int32)yy - w->yoff;
	
    dx = x2-x1;
    dy = y2-y1;  

	if(ABS(dx) > ABS(dy))
	{
		steps = ABS(dx);
	}
	else
	{
		steps = ABS(dy);
	}

	xincrement = (float)dx/(float)steps;
	yincrement = (float)dy/(float)steps;
	
	x = (float)x1;
	y = (float)y1;
	
	for(k=0;k < steps;k++)
	{
		x += xincrement;
		y += yincrement;

		/* check if we can reach the new position */
		if((float) sqrt((double)(x*x+y*y)) < (w->length-1))
			CalcNewAngel((Int32)(x-w->ex),(Int32)(y-w->ey));

		if(!w->check)
			drawjoints();
	}
	if(w->check)
		drawjoints();
}

static void* GetObjectPtr(FormPtr form, UInt16 objectID)
{
	if(form == NULL)
		form = FrmGetActiveForm();
	return(FrmGetObjectPtr(form, FrmGetObjectIndex(form, objectID)));
}
